"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import torch
import numpy as np
from sklearn import ensemble
from slingpy.models.torch_model import TorchModel
from slingpy.models.sklearn_model import SklearnModel
from typing import List, AnyStr, Union, Type, Optional
from slingpy.models.abstract_base_model import AbstractBaseModel
from genedisco.models.abstract_meta_models import AbstractMetaModel
from slingpy.models.pickleable_base_model import PickleableBaseModel
from slingpy.data_access.data_sources.abstract_data_source import AbstractDataSource
from genedisco.models.abstract_embedding_retrieval_model import EmbeddingRetrievalModel


class SklearnRandomForestRegressor(AbstractMetaModel, SklearnModel):
    """The random forest regressor model from sklean (non-deep models)"""
    def __init__(self,
                base_module: ensemble.BaseEnsemble,
                num_target_samples: Optional[int] = None):
        """The wrapping model for scikit learn ensemble regressors.

        The wrapping model augments the base_module with the capability to
        output uncertainty along with the average prediction.

        Args:
            base_module: The base forest ensemble module. It can be from 
            {RandomForestRegressor, ExtraTreesRegressor}
        """

        SklearnModel.__init__(self, base_module)
        self.n_estimators = base_module.n_estimators
        self.num_target_samples = num_target_samples

    def save(self, file_path: AnyStr):
        AbstractMetaModel.save(self, file_path)

    def save_folder(self, save_folder_path, overwrite=True):
        file_path = os.path.join(save_folder_path, "model.pickle")
        PickleableBaseModel.save(self, file_path)

    @classmethod
    def load_folder(cls: Type["TarfileSerialisationBaseModel"], save_folder_path: AnyStr) -> AbstractBaseModel:
        file_path = os.path.join(save_folder_path, "model.pickle")
        model = PickleableBaseModel.load(file_path)
        return model
    
    def get_samples(self, x: np.ndarray, k: Optional[int] = 1):
        k = self.n_estimators or k
        if k > self.n_estimators:
            raise ValueError("The number of requested samples cannot exceed"+
                             " the number of estimators in ensemble models.")
        else:
            random_indices = np.random.randint(self.n_estimators, size=k)
            y_samples = [self.model.estimators_[i].predict(x) 
                         for i in random_indices]
            y_samples = np.swapaxes(y_samples, 0, 1)
        return y_samples

    def get_model_prediction(self, data, return_std_and_margin):
        if return_std_and_margin:
            y_samples = self.get_samples(data, self.num_target_samples)
            if y_samples.ndim == 3 and y_samples.ndimsy_samples.shape[-1] > 1:
                raise NotImplementedError("Output uncertainty is only" +
                                          "implemented for 1D output.")
            else:
                y_stds = np.std(y_samples, axis=1)
                y_margins = np.max(y_samples, axis=1) - np.min(y_samples, axis=1)
                y_preds = self.model.predict(data)
                return [y_preds, y_stds, y_margins]
        else:
            y_pred = self.model.predict(data)
            return y_pred

    def predict(self,
                dataset_x: AbstractDataSource, 
                batch_size: int = 256,
                return_std_and_margin: bool = False) -> List[np.ndarray]:
        """
        Args:
            dataset_x: Input dataset to be evaluated.
            batch_size:
            return_std_and_margin: If True, return the epistemic uncertainty of the output.
        
        Returns: If return_std_and_margin is True, returns ([output_means], [output_stds]).
                        Otherwise, returns [output_means]
        """
        if self.model is None:
            self.model = self.build()

        available_indices = dataset_x.get_row_names()
        if return_std_and_margin:
            all_ids, y_preds, y_stds, y_trues, y_margins = [], [], [], [], []
        else:
            all_ids, y_preds, y_trues = [], [], []

        while len(available_indices) > 0:
            current_indices = available_indices[:batch_size]
            available_indices = available_indices[batch_size:]

            data = self.merge_strategy_x.resolve(
                dataset_x.subset(list(current_indices)).get_data())[0]

            if return_std_and_margin:
                y_pred, y_std, y_margin = self.get_model_prediction(data, return_std_and_margin=True)
                y_preds.append(y_pred)
                y_stds.append(y_std)
                y_margins.append(y_margin)
            else:
                y_pred = self.get_model_prediction(data, return_std_and_margin=False)
                y_preds.append(y_pred)
        y_preds = np.concatenate(y_preds, axis=0)
        if return_std_and_margin:
            y_stds = np.concatenate(y_stds, axis=0)
            y_margins = np.concatenate(y_margins, axis=0)
            return [y_preds, y_stds, y_margins]
        else:
            return [y_preds]

    def get_hyperopt_parameter_ranges(self):
        hyperopt_ranges = {
            "max_depth": (1, 2, 3),
            "num_estimators": (10, 50, 100, 200),
        }
        return hyperopt_ranges


class PytorchMLPRegressorWithUncertainty(AbstractMetaModel, EmbeddingRetrievalModel):
    def __init__(self,
                 model: TorchModel,
                 num_target_samples: Optional[int] = None):
        """The wrapping model for pytorch MLP regressors.

        The wrapping model augments the base_module with the capability to
        output uncertainty along with the average prediction.

        Args:
            model: The base pytorch mlp module.
        """
        super(PytorchMLPRegressorWithUncertainty, self).__init__()
        self.num_target_samples = num_target_samples
        self.model = model

    def fit(self, 
            train_x: AbstractDataSource, 
            train_y: Optional[AbstractDataSource] = None,
            validation_set_x: Optional[AbstractDataSource] = None,
            validation_set_y: Optional[AbstractDataSource] = None) -> AbstractBaseModel:
        return self.model.fit(train_x, train_y, validation_set_x, validation_set_y)

    def get_model_prediction(self,
                             data: Union[AbstractDataSource, torch.Tensor],
                             return_multiple_preds: bool=False):
        if type(data) != torch.Tensor:
            data = list(map(torch.from_numpy, data.get_data()))
        if return_multiple_preds:
            y_preds = self.get_samples(data, self.num_target_samples)
            if y_preds[0].ndim == 3 and y_preds[0].shape[-1] > 1:
                raise NotImplementedError("Output uncertainty is only" +
                                          "implemented for 1D output.")
        else:
            y_preds = self.get_samples(data, 1)
        return y_preds

    def get_samples(self, 
                    data: List[torch.Tensor], 
                    k: Optional[int] = 1) -> List[torch.Tensor]:
        y_samples = self.model.model(data, k)
        return y_samples

    def predict(self,
                dataset_x: AbstractDataSource,
                batch_size: int = 256,
                return_std_and_margin: bool = False,
                return_multiple_preds: bool = False) -> List[np.ndarray]:
        """
        Args:
            return_std_and_margin: If True, return the epistemic uncertainty of the output.
        
        Returns: If return_std is True, returns ([output_means], [output_stds]).
                        Otherwise, returns [output_means]
        """
        available_indices = dataset_x.get_row_names()
        if return_std_and_margin:
            all_ids, y_preds, y_stds, y_trues = [], [], [], []
        else:
            all_ids, y_preds, y_trues = [], [], []
        while len(available_indices) > 0:
            current_indices = available_indices[:batch_size]
            available_indices = available_indices[batch_size:]

            data = dataset_x.subset(list(current_indices))
            if return_std_and_margin:
                y_pred = self.get_model_prediction(data, return_multiple_preds=True)
                y_preds.append(y_pred)
            else:
                y_pred = self.get_model_prediction(data, return_multiple_preds=False)
                y_preds.append(y_pred)
        
        y_preds = list(
            map(lambda y_preds_i: torch.cat(y_preds_i, dim=0).detach().numpy(), 
                zip(*y_preds))
        )
        y_preds = np.squeeze(y_preds)
        if return_std_and_margin:
            y_stds = np.squeeze(np.std(y_preds, axis=1))
            y_margins = np.squeeze(
                np.max(y_preds, axis=1) - np.min(y_preds, axis=1)
            )
            y_preds = np.average(y_preds, axis=1)
            return [y_preds, y_stds, y_margins]
        else:
            return [y_preds]

    @staticmethod
    def get_submodel_file_name():
        return f"submodel.{TorchModel.get_save_file_extension()}"

    def save_folder(self, save_folder_path, overwrite=True):
        self.save_config(
            save_folder_path,
            {
                "num_target_samples": self.num_target_samples
            },
            self.get_config_file_name(),
            overwrite,
            self.__class__
        )
        model_save_path = os.path.join(save_folder_path, PytorchMLPRegressorWithUncertainty.get_submodel_file_name())
        self.model.save(model_save_path)

    @classmethod
    def load_folder(cls: Type["TarfileSerialisationBaseModel"], save_folder_path: AnyStr) -> AbstractBaseModel:
        config = cls.load_config(save_folder_path)
        instance = cls(model=None, **config)
        model_save_path = os.path.join(save_folder_path, PytorchMLPRegressorWithUncertainty.get_submodel_file_name())
        instance.model = TorchModel.load(model_save_path)
        return instance

    def get_hyperopt_parameter_ranges(self):
        hyperopt_ranges = {
            # "num_hidden_layers": (1, 2, 3),
            "hidden_layer_size": (4, 8, 16),
        }
        return hyperopt_ranges

    def get_outputs(self, dataset: AbstractDataSource, batch_size: int = 256):
        self.model.model.eval()
        current_batch, predictions, labels, hidden_representations = [], [], [], []

        def process_batch():
            x, y = zip(*current_batch)
            x, y = torch.stack(x), torch.stack(y)
            y_pred, h = self.model.model.mc_forward_impl(x, return_embedding=True)
            predictions.append(y_pred)
            labels.append(y)
            hidden_representations.append(h)

        for idx in range(len(dataset)):
            batch_data = dataset[idx]
            if isinstance(batch_data, list) and len(batch_data) == 2:
                x, y = batch_data
            else:
                x = batch_data
                y = np.zeros((1,))
            x, y = torch.from_numpy(x[0]).float(), torch.from_numpy(y)
            current_batch.append((x, y))

            if len(current_batch) >= batch_size:
                process_batch()
                current_batch = []

        if len(current_batch) != 0:
            process_batch()
        return torch.cat(predictions), torch.cat(labels)[:, 0], torch.cat(hidden_representations)

    def get_embedding(self, dataset):
        _, _, hidden_representation = self.get_outputs(dataset)
        return hidden_representation.detach()

    def get_gradient_embedding(self, dataset, eps: float = 1.0):
        y_pred, _, hidden_representation = self.get_outputs(dataset)
        return (hidden_representation * eps * torch.randn_like(y_pred.unsqueeze(-1))).detach()



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
from typing import List, AnyStr
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances
from slingpy import AbstractDataSource, AbstractBaseModel
from genedisco.active_learning_methods.acquisition_functions.base_acquisition_function import \
    BaseBatchAcquisitionFunction


class Kmeans(BaseBatchAcquisitionFunction):
    def __init__(self, representation="linear", n_init=10):
        """
            is embedding: Apply kmeans to embedding or raw data
            n_init: Specifies the number of kmeans run-throughs to use, wherein the one with the smallest inertia is
                selected for the selection phase
        """
        self.representation = representation
        self.n_init = n_init
        super(Kmeans, self).__init__()

    def __call__(self, dataset_x: AbstractDataSource, batch_size: int, available_indices: List[AnyStr],
                 last_selected_indices: List[AnyStr], last_model: AbstractBaseModel) -> List:
        if self.representation == 'linear':
            kmeans_dataset = last_model.get_embedding(dataset_x.subset(available_indices)).numpy()
        elif self.representation == 'raw':
            kmeans_dataset = np.squeeze(dataset_x.subset(available_indices), axis=1)
        else:
            raise ValueError("Representation must be one of 'linear', 'raw'")

        centers = self.kmeans_clustering(kmeans_dataset, batch_size)
        chosen = self.select_closest_to_centers(kmeans_dataset, centers)
        return [available_indices[idx] for idx in chosen]

    def kmeans_clustering(self, kmeans_dataset, num_centers):
        kmeans = KMeans(init='k-means++', n_init=self.n_init, n_clusters=num_centers).fit(kmeans_dataset)
        return kmeans.cluster_centers_

    def select_closest_to_centers(self, options, centers):
        dist_ctr = pairwise_distances(options, centers)
        min_dist_indices = np.argmin(dist_ctr, axis=0)

        return list(min_dist_indices)



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np
import pandas as pd
import slingpy as sp
from sklearn.impute import SimpleImputer
from genedisco.datasets.features.hgnc_names import HGNCNames
from slingpy.data_access.data_sources.hdf5_tools import HDF5Tools
from slingpy.data_access.data_sources.hdf5_data_source import HDF5DataSource
from slingpy.data_access.data_sources.abstract_data_source import AbstractDataSource


class CCLEProteinQuantification(object):
    """
    SOURCE: https://gygi.hms.harvard.edu/data/ccle/protein_quant_current_normalized.csv.gz
    PROJECT: https://gygi.med.harvard.edu/publications/ccle

    CITE:
    David P. Nusinow, John Szpyt, Mahmoud Ghandi, Christopher M. Rose, E. Robert McDonald III, Marian Kalocsay,
    Judit Jané-Valbuena, Ellen Gelfand, Devin K. Schweppe, Mark Jedrychowski, Javad Golji, Dale A. Porter,
    Tomas Rejtar, Y. Karen Wang, Gregory V. Kryukov, Frank Stegmeier, Brian K. Erickson, Levi A. Garraway,
     William R. Sellers, Steven P. Gygi (2020). Quantitative Proteomics of the Cancer Cell Line Encyclopedia.
     Cell 180, 2. https://doi.org/10.1016/j.cell.2019.12.02

    ACKNOWLEDGE:
    You agree to acknowledge in any work based in whole or part on the Data,
    the published paper from which the Data derives, the version of the Data, and the role of the Broad in
    its distribution. You agree to use the acknowledgement wording provided for the relevant Data in its publication.
    You will also declare in any such work that those who carried out the original analysis and collection of the Data
    bear no responsibility for the further analysis or interpretation of it.

    LICENSE: https://portals.broadinstitute.org/ccle/about
    """
    FILE_URL = "https://gygi.hms.harvard.edu/data/ccle/protein_quant_current_normalized.csv.gz"

    @staticmethod
    def load_data(save_directory) -> AbstractDataSource:
        h5_file = os.path.join(save_directory, "ccle_protein_quantification.h5")
        if not os.path.exists(h5_file):
            gz_file_path = os.path.join(save_directory, "ccle_protein_quantification.csv.gz")
            if not os.path.exists(gz_file_path):
                sp.download_streamed(CCLEProteinQuantification.FILE_URL, gz_file_path)
            df = pd.read_csv(gz_file_path, compression='gzip',
                             index_col="Gene_Symbol")
            excluded_columns = ["Protein_Id", "Description", "Group_ID", "Uniprot", "Uniprot_Acc"]
            excluded_indices = [df.columns.values.tolist().index(name) for name in excluded_columns]
            included_indices = list(sorted(set(list(range(len(df.columns)))) - set(excluded_indices)))
            data = df.values[:, included_indices].astype(float)

            si = SimpleImputer(missing_values=float("nan"), strategy='mean')
            data = si.fit_transform(data)

            row_names = df.index.values.tolist()
            name_missing_indices = np.where(list(map(lambda x: isinstance(x, float) and np.isnan(x), row_names)))[0]
            for idx in name_missing_indices:
                row_names[idx] = ""

            name_converter = HGNCNames(save_directory)
            row_names = name_converter.update_outdated_gene_names(row_names)
            row_names, idx_start = np.unique(sorted(row_names), return_index=True)
            data = data[idx_start]
            col_names = df.columns.values[included_indices].tolist()

            HDF5Tools.save_h5_file(h5_file,
                                   data,
                                   "ccle_protein_quantification",
                                   column_names=col_names,
                                   row_names=row_names)
        data_source = HDF5DataSource(h5_file, fill_missing_value=0,
                                     duplicate_merge_strategy=sp.FirstEntryMergeStrategy())
        return data_source



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import torch
import slingpy as sp
from typing import List
from torch.nn import functional as F
from genedisco.active_learning_methods.batchbald_redux import consistent_mc_dropout


class MLP(torch.nn.Module, sp.ArgumentDictionary):
    def __init__(self, input_size, hidden_size):
        super(MLP, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.fc1 = torch.nn.Linear(self.input_size, self.hidden_size * 2)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(self.hidden_size * 2, self.hidden_size)
        self.fc3 = torch.nn.Linear(self.hidden_size, 1)

    def forward(self, x):
        x = x.float()
        hidden1 = self.fc1(x)
        relu1 = self.relu(hidden1)
        hidden2 = self.fc2(relu1)
        relu2 = self.relu(hidden2)
        output = self.fc3(relu2)
        return output


class BayesianMLP(consistent_mc_dropout.BayesianModule, sp.ArgumentDictionary):
    def __init__(self, input_size: int = 808, hidden_size: int = 32):
        super(BayesianMLP, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.fc1 = torch.nn.Linear(self.input_size, self.hidden_size)
        self.fc1_drop = consistent_mc_dropout.ConsistentMCDropout2d()
        self.fc2 = torch.nn.Linear(self.hidden_size, 1)
        
    def mc_forward_impl(self, x: torch.Tensor, return_embedding=False) -> List[torch.Tensor]:
        x = x.float()
        emb = self.fc1(x)
        x = F.relu(self.fc1_drop(emb))
        x = self.fc2(x)
        if not self.training:
            x = x[:, 0]  # TODO: maybe there is a better fix to comply with the expected dimension during evaluation
        if return_embedding:
            return [x, emb]
        else:
            return [x]



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from abc import abstractmethod


class EmbeddingRetrievalModel:
    """An abstract model allowing retrieval of embedding layer and its gradient"""
    def __init__(self):
        super(EmbeddingRetrievalModel, self).__init__()

    @abstractmethod
    def get_gradient_embedding(self, dataset, eps: float = 0.001):
        raise NotImplementedError()

    @abstractmethod
    def get_embedding(self, dataset):
        raise NotImplementedError()




"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from abc import abstractmethod
from slingpy.data_access.data_sources.abstract_data_source import AbstractDataSource
from slingpy.models.tarfile_serialisation_base_model import TarfileSerialisationBaseModel


class AbstractMetaModel(TarfileSerialisationBaseModel):
    """An abstract base meta model consisting of multiple models."""
    def __init__(self):
        super(AbstractMetaModel, self).__init__()
    
    @abstractmethod
    def get_samples(self, x: AbstractDataSource, k: int = 1):
        """Generrate K sampleas [f_k(x) | k=1,2,...,K] for input x.
        
        Args:
            x: Input data of size (num_input_samples, input_dimension)
            k: The number of output values generated for each x by sampling k
                k models. k can be any intereget for Bayesian models. When
                self.base_models is a list of models, k cannot be larger than
                the length of that list. If None, make as many samples as 
                possible for ensemble models. None equals 1 for Bayesian models.
        Returns:
            y_samples: ndarray of shape (num_input_samples, k, output_dimension)
            """

        raise NotImplementedError()
    
    def get_model_prediction(self, data, return_std_and_margin):
        raise NotImplementedError



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import torch
import pickle
import slingpy as sp
from slingpy.models import torch_model
from genedisco.models import meta_models
from genedisco.models import pytorch_models
from sklearn.ensemble import RandomForestRegressor
from genedisco.datasets.features.achilles import Achilles
from typing import Any, AnyStr, Dict, List, Tuple, Union, Optional
from genedisco.datasets.features.string_embedding import STRINGEmbedding
from genedisco.datasets.screens.zhuang_2019_nk_cancer import Zhuang2019NKCancer
from genedisco.datasets.screens.schmidt_2021_t_cells_il2 import Schmidt2021TCellsIL2
from genedisco.datasets.screens.sanchez_2021_neurons_tau import Sanchez2021NeuronsTau
from genedisco.datasets.screens.schmidt_2021_t_cells_ifng import Schmidt2021TCellsIFNg
from genedisco.datasets.features.ccle_protein_quantification import CCLEProteinQuantification
from genedisco.datasets.screens.zhu_2021_sarscov2_host_factors import Zhu2021SARSCoV2HostFactors


SklearnRandomForestRegressor = meta_models.SklearnRandomForestRegressor


def update_dictionary_keys_with_prefixes(input_dict: Dict[AnyStr, Any],
                                         prefix: AnyStr):
    """Adds a prefix to the keys of a dictionary."""
    output_dict = dict(
        (prefix + key, value) for (key, value) in input_dict.items()
    )
    return output_dict


class CustomLoss(sp.TorchLoss):
    def __init__(self):
        self.loss_fn = torch.nn.MSELoss()

    def __call__(self,
                 y_pred: List[torch.Tensor],
                 y_true: List[torch.Tensor]
                 ) -> torch.Tensor:
        loss = self.loss_fn(y_pred[0], y_true[0].float())
        return loss


class SingleCycleApplication(sp.AbstractBaseApplication):
    DATASET_NAMES = [
        "shifrut_2018", "schmidt_2021_ifng",
        "schmidt_2021_il2", "zhuang_2019_nk",
        "sanchez_2021_tau", "zhu_2021_sarscov2"
    ]

    FEATURE_SET_NAMES = ["achilles", "ccle", "string"]

    def __init__(
        self,
        dataset_name: AnyStr = DATASET_NAMES[0],
        feature_set_name: AnyStr = FEATURE_SET_NAMES[0],
        output_directory: AnyStr = "",
        cache_directory: AnyStr = "",
        schedule_on_slurm: bool = False,
        remote_execution_time_limit_days: int = 3,
        remote_execution_mem_limit_in_mb: int = 2048,
        remote_execution_virtualenv_path: AnyStr = "",
        remote_execution_num_cpus: int = 1,
        split_index_outer: int = 0,
        split_index_inner: int = 0,
        num_splits_outer: int = 2,
        num_splits_inner: int = 2,
        model_name: AnyStr = "mlp",
        evaluate_against: AnyStr = "test",
        selected_indices_file_path: AnyStr = "",
        test_indices_file_path: AnyStr = "",
        train_ratio: float = 0.8,
        hyperopt: bool = False,
        num_hyperopt_runs: int = 15,
        hyperopt_offset: int = 0,
        hyperopt_metric_name: AnyStr = "MeanAbsoluteError",
        train: bool = True,
        rf_max_depth: int = -1,  # randomforest hyperparams
        rf_num_estimators: int = 100,  # ensemble_model_hyperparms
        dn_num_layers: int = 2,  # deep net hyperparams
        dn_hidden_layer_size: int = 16,  # deep net hyperparams
        seed: int = 0
    ):
        model_hyperparams = {
            "rf_max_depth": rf_max_depth,
            "rf_num_estimators": rf_num_estimators,
            "dn_num_layers": dn_num_layers,
            "dn_hidden_layer_size": dn_hidden_layer_size,
        }
        self.rf_max_depth = rf_max_depth
        self.rf_num_estimators = rf_num_estimators
        self.dn_num_layers = dn_num_layers
        self.dn_hidden_layer_size = dn_hidden_layer_size
        self.model_name = model_name
        self.train = train
        self.cache_directory = cache_directory
        self.test_indices_file_path = test_indices_file_path
        self.selected_indices_file_path = selected_indices_file_path
        self.train_ratio = train_ratio
        self.dataset_name = dataset_name
        self.feature_set_name = feature_set_name
        self.hyperopt_offset = hyperopt_offset
        self.hyperopt_metric_name = hyperopt_metric_name
        self.seed = seed
        self.model = self.get_model(model_name, **model_hyperparams)

        with open(self.test_indices_file_path, "rb") as fp:
            self.test_indices = pickle.load(fp)
        with open(self.selected_indices_file_path, "rb") as fp:
            self.selected_indices = pickle.load(fp)

        super(SingleCycleApplication, self).__init__(
            output_directory=output_directory,
            schedule_on_slurm=schedule_on_slurm,
            split_index_outer=split_index_outer,
            split_index_inner=split_index_inner,
            num_splits_outer=num_splits_outer,
            num_splits_inner=num_splits_inner,
            evaluate_against=evaluate_against,
            nested_cross_validation=False,
            save_predictions=False,
            save_predictions_file_format="tsv",
            hyperopt=hyperopt,
            num_hyperopt_runs=num_hyperopt_runs,
            seed=seed,
            remote_execution_time_limit_days=remote_execution_time_limit_days,
            remote_execution_virtualenv_path=remote_execution_virtualenv_path,
            remote_execution_num_cpus=remote_execution_num_cpus,
            remote_execution_mem_limit_in_mb=remote_execution_mem_limit_in_mb
        )

    def get_metrics(self, set_name: AnyStr) -> List[sp.AbstractMetric]:
        return [
            sp.metrics.MeanAbsoluteError(),
            sp.metrics.RootMeanSquaredError(),
            sp.metrics.SymmetricMeanAbsolutePercentageError(),
            sp.metrics.SpearmanRho()
        ]

    @staticmethod
    def get_dataset_y(dataset_name, cache_directory):
        if dataset_name == SingleCycleApplication.DATASET_NAMES[0]:
            # dataset_y = Shifrut2018TCells.load_data(cache_directory)
            raise NotImplementedError("Shifrut et al is currently not available for automated evaluation.")
        elif dataset_name == SingleCycleApplication.DATASET_NAMES[1]:
            dataset_y = Schmidt2021TCellsIFNg.load_data(cache_directory)
        elif dataset_name == SingleCycleApplication.DATASET_NAMES[2]:
            dataset_y = Schmidt2021TCellsIL2.load_data(cache_directory)
        elif dataset_name == SingleCycleApplication.DATASET_NAMES[3]:
            dataset_y = Zhuang2019NKCancer.load_data(cache_directory)
        elif dataset_name == SingleCycleApplication.DATASET_NAMES[4]:
            dataset_y = Sanchez2021NeuronsTau.load_data(cache_directory)
        elif dataset_name == SingleCycleApplication.DATASET_NAMES[5]:
            dataset_y = Zhu2021SARSCoV2HostFactors.load_data(cache_directory)
        else:
            raise NotImplementedError(f"{dataset_name} is not implemented.")
        return dataset_y

    @staticmethod
    def get_dataset_x(feature_set_name, cache_directory):
        if feature_set_name == SingleCycleApplication.FEATURE_SET_NAMES[0]:
            dataset = Achilles.load_data(cache_directory)
        elif feature_set_name == SingleCycleApplication.FEATURE_SET_NAMES[1]:
            dataset = CCLEProteinQuantification.load_data(cache_directory)
        elif feature_set_name == SingleCycleApplication.FEATURE_SET_NAMES[2]:
            dataset = STRINGEmbedding.load_data(cache_directory)
        else:
            raise NotImplementedError()

        dataset_x = sp.CompositeDataSource([dataset])
        return dataset_x

    def load_data(self) -> Dict[AnyStr, sp.AbstractDataSource]:
        dataset_x = SingleCycleApplication.get_dataset_x(self.feature_set_name, self.cache_directory)
        dataset_y = SingleCycleApplication.get_dataset_y(self.dataset_name, self.cache_directory)

        # Subset dataset_y by the overlap of genes present in both dataset_x and dataset_y.
        avail_names = sorted(
            list(set(dataset_x.get_row_names()).intersection(
                 set(dataset_y.get_row_names())))
        )
        dataset_y = dataset_y.subset(avail_names)
        dataset_x = dataset_x.subset(avail_names)

        dataset_y = dataset_y.subset(self.selected_indices)

        stratifier = sp.StratifiedSplit()
        training_indices, validation_indices = stratifier.split(
            dataset_y,
            test_set_fraction=1 - self.train_ratio,
            split_index=self.split_index_inner
        )
        return {
            "training_set_x": dataset_x.subset(training_indices),
            "training_set_y": dataset_y.subset(training_indices),
            "validation_set_x": dataset_x.subset(validation_indices),
            "validation_set_y": dataset_y.subset(validation_indices),
            "test_set_x": dataset_x.subset(self.test_indices),
            "test_set_y": dataset_y.subset(self.test_indices)
        }

    def get_model(self, model_name, **kwargs):
        if model_name == "randomforest":
            rf_max_depth = kwargs["rf_max_depth"]
            if rf_max_depth == -1:
                rf_max_depth = None
            sp_model = SklearnRandomForestRegressor(
                base_module=RandomForestRegressor(
                    n_estimators=kwargs["rf_num_estimators"],
                    max_depth=rf_max_depth,
                    random_state=self.seed)
            )
        elif model_name == "bayesian_mlp":
            super_base_module = torch_model.TorchModel(
                base_module=pytorch_models.BayesianMLP(
                    input_size=SingleCycleApplication.get_dataset_x(self.feature_set_name,
                                                                    self.cache_directory).get_shape()[0][-1],
                    hidden_size=self.dn_hidden_layer_size),
                loss=CustomLoss(),
                batch_size=64,
                num_epochs=100
            )
            sp_model = meta_models.PytorchMLPRegressorWithUncertainty(
                model=super_base_module,
                num_target_samples=100
            )
        else:
            raise NotImplementedError(f"{model_name} does not exist.")
        return sp_model

    def train_model(self) -> Optional[sp.AbstractBaseModel]:
        self.model.fit(self.datasets.training_set_x,
                       self.datasets.training_set_y)
        return self.model

    def get_hyperopt_parameter_ranges(self) -> Dict[AnyStr, Union[List, Tuple]]:
        """
        Get hyper-parameter optimization ranges.

        Returns:
            A dictionary with each item corresponding to a named hyper-parameter and its associated discrete
            (represented as a Tuple) or continuous (represented as a List[start, end]) value range.
        """
        model_hyperopt_parameter_ranges = self.get_model_hyperparameter_ranges()
        hyperopt_ranges = {}
        hyperopt_ranges.update(model_hyperopt_parameter_ranges)
        return hyperopt_ranges

    def get_model_hyperparameter_ranges(self) -> Dict[AnyStr, Union[AnyStr, int]]:
        """
        Get the hyper-parameter optimization ranges for the evaluated model.

        Returns:
            A dictionary with each item corresponding to a named hyper-parameter and its associated discrete
            (represented as a Tuple) or continuous (represented as a List[start, end]) value range.
        """
        prefixes = {"randomforest": "rf_", "bayesian_mlp": "dn_"}
        model_hyperopt_parameter_ranges = self.model.get_hyperopt_parameter_ranges()
        model_hyperopt_parameter_ranges = update_dictionary_keys_with_prefixes(
            model_hyperopt_parameter_ranges, prefixes[self.model_name]
        )
        return model_hyperopt_parameter_ranges


if __name__ == "__main__":
    app = sp.instantiate_from_command_line(SingleCycleApplication)
    results = app.run()






"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np
import slingpy as sp
from functools import partial
from slingpy.utils.path_tools import PathTools
from typing import AnyStr, List, Optional, Dict
from slingpy.utils.nestable_pool import NestablePool as Pool
from genedisco.apps.active_learning_loop import ActiveLearningLoop
from slingpy import AbstractMetric, AbstractBaseModel, AbstractDataSource
from genedisco.apps.single_cycle_application import SingleCycleApplication


class RunExperimentsApplication(sp.AbstractBaseApplication):
    def __init__(self,
                 output_directory: AnyStr,
                 cache_directory: AnyStr,
                 model_name: AnyStr = "bayesian_mlp",
                 acquisition_function_path: AnyStr = None,
                 max_num_jobs: int = 90,
                 seed: int = 909,
                 hyperopt_children: bool = True,
                 num_random_seeds: int = 5,
                 num_active_learning_cycles: int = 20,
                 acquisition_batch_size: int = 256,
                 schedule_on_slurm: bool = False,
                 schedule_children_on_slurm: bool = False,
                 remote_execution_time_limit_days: int = 5,
                 remote_execution_mem_limit_in_mb: int = 8048,
                 remote_execution_virtualenv_path: AnyStr = "",
                 remote_execution_num_cpus: int = 1):
        self.cache_directory = cache_directory
        self.model_name = model_name
        self.acquisition_function_path = acquisition_function_path
        self.hyperopt_children = hyperopt_children
        self.schedule_children_on_slurm = schedule_children_on_slurm
        self.max_num_jobs = max_num_jobs
        self.num_random_seeds = num_random_seeds
        self.num_active_learning_cycles = num_active_learning_cycles
        self.acquisition_batch_size = acquisition_batch_size
        super(RunExperimentsApplication, self).__init__(
            output_directory=output_directory,
            evaluate=False,
            hyperopt=False,
            single_run=True,
            save_predictions=False,
            seed=seed,
            schedule_on_slurm=schedule_on_slurm,
            remote_execution_time_limit_days=remote_execution_time_limit_days,
            remote_execution_mem_limit_in_mb=remote_execution_mem_limit_in_mb,
            remote_execution_virtualenv_path=remote_execution_virtualenv_path,
            remote_execution_num_cpus=remote_execution_num_cpus
        )

    def load_data(self) -> Dict[AnyStr, AbstractDataSource]:
        return {}

    def get_metrics(self, set_name: AnyStr) -> List[AbstractMetric]:
        return []

    @staticmethod
    def parallel_run_wrapper(args, self_reference):
        seed, baseline, acqfunc_path, dataset, feature_set, model_output_directory = args
        app = ActiveLearningLoop(seed=seed,
                                 num_active_learning_cycles=self_reference.num_active_learning_cycles,
                                 acquisition_batch_size=self_reference.acquisition_batch_size,
                                 acquisition_function_name=baseline,
                                 acquisition_function_path = acqfunc_path,
                                 hyperopt_children=self_reference.hyperopt_children,
                                 dataset_name=dataset,
                                 feature_set_name=feature_set,
                                 schedule_on_slurm=self_reference.schedule_children_on_slurm,
                                 schedule_children_on_slurm=self_reference.schedule_children_on_slurm,
                                 remote_execution_time_limit_days=self_reference.remote_execution_time_limit_days,
                                 remote_execution_mem_limit_in_mb=self_reference.remote_execution_mem_limit_in_mb,
                                 remote_execution_num_cpus=self_reference.remote_execution_num_cpus,
                                 remote_execution_virtualenv_path=self_reference.remote_execution_virtualenv_path,
                                 model_name=self_reference.model_name,
                                 output_directory=model_output_directory,
                                 cache_directory=self_reference.cache_directory)
        app.run()
        return None

    def train_model(self) -> Optional[AbstractBaseModel]:
        acqfunc_path = self.acquisition_function_path
        random_state = np.random.RandomState(self.seed)
        baselines = ActiveLearningLoop.ACQUISITION_FUNCTIONS
        random_seeds = random_state.randint(2**31, size=self.num_random_seeds)
        datasets = SingleCycleApplication.DATASET_NAMES[1:]  # Excluding Shifrut et al.
        feature_sets = SingleCycleApplication.FEATURE_SET_NAMES

        arg_list = []
        for dataset in datasets:
            dataset_output_directory = os.path.join(self.output_directory, f"data_{dataset}")
            PathTools.mkdir_if_not_exists(dataset_output_directory)

            for feature_set in feature_sets:
                feature_output_directory = os.path.join(dataset_output_directory, f"feat_{feature_set}")
                PathTools.mkdir_if_not_exists(feature_output_directory)

                for baseline in baselines:
                    baseline_output_directory = os.path.join(feature_output_directory, f"acq_{baseline}")
                    PathTools.mkdir_if_not_exists(baseline_output_directory)

                    for seed in random_seeds:
                        model_output_directory = os.path.join(baseline_output_directory, f"seed_{seed}")
                        arg_list.append((seed, baseline, acqfunc_path, dataset, feature_set, model_output_directory))

        max_num_processes = self.max_num_jobs
        num_processes = min(len(arg_list), max_num_processes)
        if num_processes == 1:
            outputs = []
            for arg in arg_list:
                outputs.append(RunExperimentsApplication.parallel_run_wrapper(arg, self_reference=self))
        else:
            with Pool(processes=num_processes) as pool:
                outputs = list(pool.imap_unordered(
                    partial(RunExperimentsApplication.parallel_run_wrapper, self_reference=self),
                    arg_list, chunksize=1))
        return None


def main():
    app = sp.instantiate_from_command_line(RunExperimentsApplication)
    results = app.run()


if __name__ == "__main__":
    main()



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
from typing import AnyStr, List
from slingpy import AbstractDataSource
from slingpy.models.abstract_base_model import AbstractBaseModel
from genedisco.active_learning_methods.acquisition_functions.base_acquisition_function import \
    BaseBatchAcquisitionFunction


class MarginSamplingAcquisition(BaseBatchAcquisitionFunction):
    def __call__(self,
                 dataset_x: AbstractDataSource,
                 select_size: int,
                 available_indices: List[AnyStr],
                 last_selected_indices: List[AnyStr] = None,
                 model: AbstractBaseModel = None,
                 ) -> List:
        avail_dataset_x = dataset_x.subset(available_indices)
        model_pedictions = model.predict(avail_dataset_x, return_std_and_margin=True)

        if len(model_pedictions) != 3:
            raise TypeError("The provided model does not output margins.")

        # print("Hi! This is Margin Sampling!")
        pred_mean, pred_uncertainties, pred_margins = model_pedictions
        # print(pred_margins)

        if len(pred_mean) < select_size:
            raise ValueError("The number of query samples exceeds"
                             "the size of the available data.")

        numerical_selected_indices = np.flip(
            np.argsort(pred_margins)
        )[:select_size]
        selected_indices = [available_indices[i] for i in numerical_selected_indices]

        return selected_indices



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from abc import abstractmethod
from typing import AnyStr, List
from slingpy import AbstractDataSource
from slingpy.models.abstract_base_model import AbstractBaseModel


class BaseBatchAcquisitionFunction(object):
    @abstractmethod
    def __call__(self,
                 dataset_x: AbstractDataSource,
                 batch_size: int,
                 available_indices: List[AnyStr],
                 last_selected_indices: List[AnyStr],
                 last_model: AbstractBaseModel,
                 ) -> List:
        """
        Nominate experiments for the next learning round.

        Args:
            dataset_x: The dataset containing all training samples.
            batch_size: Size of the batch to acquire.
            available_indices: The list of the indices (names) of the samples not
                chosen in the previous rounds.
            last_selected_indices: The set of indices selected in the prrevious
                cycle.
            last_model: The prediction model trained by labeled samples chosen so far.

        Returns:
            A list of indices (names) of the samples chosen for the next round.
        """
        raise NotImplementedError()



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np
import pandas as pd
import slingpy as sp
from typing import AnyStr
from slingpy.utils.logging import warn


class HGNCNames(object):
    FILE_URL = "http://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/tsv/locus_types/gene_with_protein_product.txt"

    def __init__(self, cache_directory: AnyStr):
        self.hgnc_mappings = {}
        self.cache_directory = cache_directory

    def get_hgnc_master_file(self):
        tsv_file = os.path.join(self.cache_directory, "hgnc_mapping.tsv")
        if not os.path.exists(tsv_file):
            sp.download_streamed(HGNCNames.FILE_URL, tsv_file)
        return tsv_file

    def get_gene_names(self, index_name: AnyStr = "symbol"):
        tsv_file = self.get_hgnc_master_file()
        data = pd.read_csv(tsv_file, sep="\t", index_col=index_name)
        gene_names = data.index.values.tolist()
        return gene_names

    def get_hgnc_mapping(self, from_id: AnyStr = "uniprot_ids", to_id: AnyStr = "symbol",
                         split_symbol: AnyStr = "|"):
        cache_entry_name = f"{from_id}${to_id}"
        if cache_entry_name not in self.hgnc_mappings:
            tsv_file = self.get_hgnc_master_file()
            mapping_data = pd.read_csv(tsv_file, sep="\t", index_col=from_id)
            target_column_index = mapping_data.columns.values.tolist().index(to_id)
            mapping_row_names = mapping_data.index.values
            mapping_data = mapping_data.values[:, target_column_index]

            if from_id in ["prev_symbol", "alias_symbol"]:
                new_mapping_data, new_mapping_row_names = [], []
                for idx, row_name in enumerate(mapping_row_names):
                    if isinstance(row_name, float) and np.isnan(row_name):
                        continue

                    split_names = row_name.split(split_symbol)
                    for split_name in split_names:
                        if split_name == "":
                            continue
                        new_mapping_data.append(mapping_data[idx])
                        new_mapping_row_names.append(split_name)
                mapping_data = np.array(new_mapping_data)
                mapping_row_names = new_mapping_row_names

            assert len(mapping_row_names) == len(mapping_data)

            circular_entries = []
            mapping = dict(zip(mapping_row_names, list(map(str, mapping_data))))
            for from_id, to_id in mapping.items():
                cur_to_id = to_id
                circle_detected, previous_stack = False, {cur_to_id}
                while cur_to_id in mapping:
                    cur_to_id = mapping[cur_to_id]
                    if cur_to_id in previous_stack:
                        circle_detected = True
                        break
                    else:
                        previous_stack.add(cur_to_id)
                if circle_detected:
                    circular_entries.append(to_id)
                elif cur_to_id != to_id:  # Replace with final node in graph.
                    mapping[from_id] = cur_to_id

            for entry in set(circular_entries):
                mapping[entry] = entry
            self.hgnc_mappings[cache_entry_name] = mapping
        return self.hgnc_mappings[cache_entry_name]

    def update_outdated_gene_names(self, row_names, verbose: bool = False):
        row_names = list(row_names)
        for other_id_name in ["prev_symbol", "alias_symbol"]:
            previous_mapping = self.get_hgnc_mapping(to_id="symbol", from_id=other_id_name)
            for idx, row_name in enumerate(row_names):
                before = row_name
                if row_name in previous_mapping:
                    row_names[idx] = previous_mapping[row_name]
                if before != row_names[idx] and verbose:
                    warn(f"{before} => {row_names[idx]}")
        return row_names



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from distutils.core import setup
from setuptools import find_packages

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

setup(
    name='genedisco',
    version='1.0.0',
    python_requires=">=3.8",
    packages=find_packages(),
    package_data={
        "": ["*.txt"],
        "": ["*.csv"],
    },
    author='see README.txt',
    url="https://gsk.ai/genedisco-challenge/",
    author_email='patrick.x.schwab@gsk.com',
    license="Apache-2.0",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'run_experiments=genedisco.apps.run_experiments_application:main',
            'active_learning_loop=genedisco.apps.active_learning_loop:main'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ]
)



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import torch
import numpy as np
from typing import List, AnyStr
from slingpy import AbstractDataSource, AbstractBaseModel
from genedisco.active_learning_methods.acquisition_functions.base_acquisition_function import \
    BaseBatchAcquisitionFunction

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class AdversarialBIM(BaseBatchAcquisitionFunction):
    def __init__(self, args=None):
        if args is None:
            args = {}

        if 'eps' in args:
            self.eps = args['eps']
        else:
            self.eps = 0.05

        if 'verbose' in args:
            self.verbose = args['verbose']
        else:
            self.verbose = True

        if 'stop_iterations_by_count' in args:
            self.stop_iterations_by_count = args['stop_iterations_by_count']
        else:
            self.stop_iterations_by_count = 1000

        if 'gamma' in args:
            self.gamma = args['gamma']
        else:
            self.gamma = 0.35

        if 'adversarial_sample_ratio' in args:
            self.adversarial_sample_ratio = args['adversarial_sample_ratio']
        else:
            self.adversarial_sample_ratio = 0.1

        super(AdversarialBIM, self).__init__()

    def __call__(self, dataset_x: AbstractDataSource, batch_size: int, available_indices: List[AnyStr],
                 last_selected_indices: List[AnyStr], last_model: AbstractBaseModel) -> List:
        dis = np.zeros(len(available_indices)) + np.inf
        data_pool = dataset_x.subset(available_indices)

        for i, index in enumerate(available_indices[:100]):
            x = torch.as_tensor(data_pool.subset([index]).get_data()).to(device)
            dis[i] = self.cal_dis(x, last_model)

        chosen = dis.argsort()[:batch_size]
        for x in np.sort(dis)[:batch_size]:
            print(x)
        return [available_indices[idx] for idx in chosen]

    def cal_dis(self, x, last_model):
        nx = x.detach()
        first_x = torch.clone(nx)

        nx.requires_grad_()
        eta = torch.zeros(nx.shape).to(device)
        iteration = 0

        while torch.linalg.norm(nx + eta - first_x) < self.gamma * torch.linalg.norm(first_x):

            if iteration >= self.stop_iterations_by_count:
                break

            out = torch.as_tensor(last_model.get_model_prediction(nx + eta, return_multiple_preds=True)[0])
            out = torch.squeeze(out)
            variance = torch.var(out)
            variance.backward()

            eta += self.eps * torch.sign(nx.grad.data)
            nx.grad.data.zero_()
            iteration += 1
        return (eta * eta).sum()



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import sys
import pickle
import inspect
import numpy as np
import slingpy as sp
import importlib.util
from slingpy.utils.path_tools import PathTools
from typing import AnyStr, Dict, List, Optional
from slingpy import AbstractMetric, AbstractBaseModel, AbstractDataSource
from genedisco.active_learning_methods.acquisition_functions.kmeans import Kmeans
from genedisco.active_learning_methods.acquisition_functions.core_set import CoreSet
from genedisco.apps.single_cycle_application import SingleCycleApplication, CustomLoss
from genedisco.active_learning_methods.acquisition_functions.badge_sampling import BadgeSampling
from genedisco.active_learning_methods.acquisition_functions.adversarial_bim import AdversarialBIM
from genedisco.active_learning_methods.acquisition_functions.uncertainty_acquisition import TopUncertainAcquisition
from genedisco.active_learning_methods.acquisition_functions.uncertainty_acquisition import SoftUncertainAcquisition
from genedisco.active_learning_methods.acquisition_functions.margin_sampling_acquisition import \
    MarginSamplingAcquisition
from genedisco.active_learning_methods.acquisition_functions.base_acquisition_function import \
    BaseBatchAcquisitionFunction
from genedisco.active_learning_methods.acquisition_functions.random_acquisition_function import \
    RandomBatchAcquisitionFunction


class ActiveLearningLoop(sp.AbstractBaseApplication):
    ACQUISITION_FUNCTIONS = [
        "random", "topuncertain", "softuncertain", "marginsample", "coreset", "badge",
        "kmeans_embedding", "kmeans_data", "adversarialBIM", "custom"
    ]

    def __init__(
        self,
        model_name: AnyStr = "randomforest",
        acquisition_function_name: AnyStr = "random",
        acquisition_function_path: AnyStr = "custom",
        acquisition_batch_size: int = 32,
        num_active_learning_cycles: int = 10,
        feature_set_name: AnyStr = SingleCycleApplication.FEATURE_SET_NAMES[0],
        dataset_name: AnyStr = SingleCycleApplication.DATASET_NAMES[1],
        cache_directory: AnyStr = "",
        output_directory: AnyStr = "",
        test_ratio: float = 0.2,
        hyperopt_children: bool = True,
        schedule_on_slurm: bool = False,
        schedule_children_on_slurm: bool = False,
        remote_execution_time_limit_days: int = 1,
        remote_execution_mem_limit_in_mb: int = 2048,
        remote_execution_virtualenv_path: AnyStr = "",
        remote_execution_num_cpus: int = 1,
        seed: int = 0
    ):
        self.acquisition_function_name = acquisition_function_name
        self.acquisition_function_path = acquisition_function_path
        PathTools.mkdir_if_not_exists(output_directory)
        self.acquisition_function = ActiveLearningLoop.get_acquisition_function(
            self.acquisition_function_name,
            self.acquisition_function_path
        )
        self.acquisition_batch_size = acquisition_batch_size
        self.num_active_learning_cycles = num_active_learning_cycles
        self.dataset_name = dataset_name
        self.feature_set_name = feature_set_name
        self.model_name = model_name
        self.hyperopt_children = hyperopt_children
        self.test_ratio = test_ratio
        self.cache_directory = cache_directory
        self.schedule_children_on_slurm = schedule_children_on_slurm
        super(ActiveLearningLoop, self).__init__(
            output_directory=output_directory,
            seed=seed,
            evaluate=False,
            hyperopt=False,
            single_run=True,
            save_predictions=False,
            schedule_on_slurm=schedule_on_slurm,
            remote_execution_num_cpus=remote_execution_num_cpus,
            remote_execution_time_limit_days=remote_execution_time_limit_days,
            remote_execution_mem_limit_in_mb=remote_execution_mem_limit_in_mb,
            remote_execution_virtualenv_path=remote_execution_virtualenv_path
        )
    
    @staticmethod
    def get_acquisition_function(
            acquisition_function_name: AnyStr,
            acquisition_function_path: AnyStr
    ) -> BaseBatchAcquisitionFunction:
        if acquisition_function_name == "random":
            return RandomBatchAcquisitionFunction()
        elif acquisition_function_name == "topuncertain":
            return TopUncertainAcquisition()
        elif acquisition_function_name == "softuncertain":
            return SoftUncertainAcquisition()
        elif acquisition_function_name == "marginsample":
            return MarginSamplingAcquisition()
        elif acquisition_function_name == "badge":
            return BadgeSampling()
        elif acquisition_function_name == "coreset":
            return CoreSet()
        elif acquisition_function_name == "kmeans_embedding":
            return Kmeans(representation="linear", n_init=10)
        elif acquisition_function_name == "kmeans_data":
            return Kmeans(representation="raw", n_init=10)
        elif acquisition_function_name == "adversarialBIM":
            return AdversarialBIM()
        elif acquisition_function_name == "custom":
            acqfunc_class = ActiveLearningLoop.get_if_valid_acquisition_function_file(acquisition_function_path)
            return acqfunc_class()
        else:
            raise NotImplementedError()

    @staticmethod
    def get_if_valid_acquisition_function_file(acquisition_function_path: AnyStr):
        if not os.path.exists(acquisition_function_path):
            raise ValueError("The path to the acquisition function file does not exist.")
        else:
            module_name = "custom_acqfunc"
            spec = importlib.util.spec_from_file_location(module_name, acquisition_function_path)
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
            custom_acqfunc_class = None
            for name, obj in inspect.getmembers(module, inspect.isclass):
                obj_bases = obj.__bases__
                if BaseBatchAcquisitionFunction in obj_bases:
                    custom_acqfunc_class = obj
            if custom_acqfunc_class is None:
                raise ValueError(f"No valid acquisition function was found at {acquisition_function_path}. "
                                 f"Did you forget to inherit from 'BaseBatchAcquisitionFunction'?")
            return custom_acqfunc_class

    def initialize_pool(self):
        dataset_x = SingleCycleApplication.get_dataset_x(
            self.feature_set_name,
            self.cache_directory
        )
        dataset_y = SingleCycleApplication.get_dataset_y(
            self.dataset_name, 
            self.cache_directory
        )
        available_indices = sorted(
            list(set(dataset_x.get_row_names()).intersection(set(dataset_y.get_row_names())))
        )
        test_indices = sorted(
            list(
                np.random.choice(
                    available_indices, 
                    size=int(self.test_ratio * len(available_indices)), 
                    replace=False)
            )
        )
        available_indices = list(set(available_indices) - set(test_indices))
        return dataset_x, available_indices, test_indices

    def load_data(self) -> Dict[AnyStr, AbstractDataSource]:
        return {}

    def get_metrics(self, set_name: AnyStr) -> List[AbstractMetric]:
        return []

    def train_model(self) -> Optional[AbstractBaseModel]:
        single_cycle_application_args = {
            "model_name": self.model_name,
            "seed": self.seed,
            "remote_execution_time_limit_days": self.remote_execution_time_limit_days,
            "remote_execution_mem_limit_in_mb": self.remote_execution_mem_limit_in_mb,
            "remote_execution_virtualenv_path": self.remote_execution_virtualenv_path,
            "remote_execution_num_cpus": self.remote_execution_num_cpus,
            "schedule_on_slurm": self.schedule_children_on_slurm,
        }
        cumulative_indices = []
        dataset_x, available_indices, test_indices = self.initialize_pool()

        last_selected_indices = sorted(
            list(
                np.random.choice(available_indices, 
                                 size=int(self.acquisition_batch_size),
                                 replace=False)
            )
        )
        cumulative_indices += last_selected_indices
        result_records = list()
        for cycle_index in range(self.num_active_learning_cycles):
            current_cycle_directory = os.path.join(self.output_directory, f"cycle_{cycle_index}")
            PathTools.mkdir_if_not_exists(current_cycle_directory)

            cumulative_indices_file_path = os.path.join(current_cycle_directory, "selected_indices.pickle")
            with open(cumulative_indices_file_path, "wb") as fp:
                pickle.dump(cumulative_indices, fp)
            test_indices_file_path = os.path.join(current_cycle_directory, "test_indices.pickle")
            with open(test_indices_file_path, "wb") as fp:
                pickle.dump(test_indices, fp)

            app = SingleCycleApplication(
                dataset_name=self.dataset_name,
                feature_set_name=self.feature_set_name,
                cache_directory=self.cache_directory,
                output_directory=current_cycle_directory,
                train_ratio=0.8,
                hyperopt=self.hyperopt_children,
                selected_indices_file_path=cumulative_indices_file_path,
                test_indices_file_path=test_indices_file_path,
                **single_cycle_application_args
            )
            results = app.run().run_result
            result_records.append(results.test_scores)
            available_indices = list(
                set(available_indices) - set(last_selected_indices)
            )

            trained_model_path = results.model_path
            trained_model = app.model.load(trained_model_path)

            last_selected_indices = self.acquisition_function(
                dataset_x,
                self.acquisition_batch_size,
                available_indices,
                last_selected_indices,
                trained_model
            )
            cumulative_indices.extend(last_selected_indices)
            cumulative_indices = list(set(cumulative_indices))
            assert len(last_selected_indices) == self.acquisition_batch_size

        results_path = os.path.join(self.output_directory, "results.pickle")
        with open(results_path, "wb") as fp:
            pickle.dump(result_records, fp)
        return None


def main():
    active_learning_loop = sp.instantiate_from_command_line(ActiveLearningLoop)
    active_learning_loop.run()


if __name__ == "__main__":
    main()



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np
import pandas as pd
import slingpy as sp
from genedisco.datasets.features.hgnc_names import HGNCNames
from slingpy.data_access.data_sources.hdf5_tools import HDF5Tools
from slingpy.data_access.data_sources.hdf5_data_source import HDF5DataSource
from slingpy.data_access.data_sources.abstract_data_source import AbstractDataSource


class STRINGEmbedding(object):
    """
    SOURCE:
    https://groups.csail.mit.edu/cb/mashup/vectors/string_human_genes.txt
    https://groups.csail.mit.edu/cb/mashup/vectors/string_human_mashup_vectors_d800.txt

    PROJECT: http://cb.csail.mit.edu/cb/mashup/

    CITE:
    https://string-db.org/cgi/about?sessionId=bGpYHQrYvsOz&footer_active_subpage=references

    ACKNOWLEDGE:
    https://string-db.org/cgi/about?sessionId=bGpYHQrYvsOz&footer_active_subpage=references

    LICENSE: https://creativecommons.org/licenses/by/4.0/
    """
    FILE_URL = "https://groups.csail.mit.edu/cb/mashup/vectors/string_human_mashup_vectors_d800.txt"
    NAME_FILE_URL = "https://groups.csail.mit.edu/cb/mashup/vectors/string_human_genes.txt"

    @staticmethod
    def load_data(save_directory) -> AbstractDataSource:
        h5_file = os.path.join(save_directory, "string_embedding.h5")
        if not os.path.exists(h5_file):
            csv_file_path = os.path.join(save_directory, "string_human_mashup_vectors_d800.txt")
            if not os.path.exists(csv_file_path):
                sp.download_streamed(STRINGEmbedding.FILE_URL, csv_file_path)
            name_file_path = os.path.join(save_directory, "string_human_genes.txt")
            if not os.path.exists(name_file_path):
                sp.download_streamed(STRINGEmbedding.NAME_FILE_URL, name_file_path)

            df = pd.read_csv(csv_file_path, index_col=None, header=None, sep="\t")
            row_names = pd.read_csv(name_file_path, index_col=None, header=None).values[:, 0].tolist()

            data = df.values[:, 1:]
            name_converter = HGNCNames(save_directory)
            row_names = name_converter.update_outdated_gene_names(row_names)
            row_names, idx_start = np.unique(sorted(row_names), return_index=True)
            data = data[idx_start]
            col_names = [f"feature_{idx}" for idx in range(799)]

            HDF5Tools.save_h5_file(h5_file,
                                   data,
                                   "string_embedding",
                                   column_names=col_names,
                                   row_names=row_names)
        data_source = HDF5DataSource(h5_file, fill_missing_value=0,
                                     duplicate_merge_strategy=sp.FirstEntryMergeStrategy())
        return data_source



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np
import pandas as pd
import slingpy as sp
from sklearn.impute import SimpleImputer
from genedisco.datasets.features.hgnc_names import HGNCNames
from slingpy.data_access.data_sources.hdf5_tools import HDF5Tools
from slingpy.data_access.data_sources.hdf5_data_source import HDF5DataSource
from slingpy.data_access.data_sources.abstract_data_source import AbstractDataSource


class Achilles(object):
    """
    Data from:
    SOURCE: https://ndownloader.figshare.com/files/25494359
    PROJECT: https://depmap.org/portal/download/

    CITE:
    DepMap, Broad (2020): DepMap 20Q4 Public. figshare. Dataset doi:10.6084/m9.figshare.13237076.v1.
    Robin M. Meyers, Jordan G. Bryan, James M. McFarland, Barbara A. Weir, ... David E. Root, William C. Hahn,
    Aviad Tsherniak. Computational correction of copy number effect improves specificity of CRISPR-Cas9 essentiality
    screens in cancer cells. Nature Genetics 2017 October 49:1779–1784. doi:10.1038/ng.3984
    Dempster, J. M., Rossen, J., Kazachkova, M., Pan, J., Kugener, G., Root, D. E., & Tsherniak, A. (2019).
    Extracting Biological Insights from the Project Achilles Genome-Scale CRISPR Screens in Cancer Cell Lines.
    BioRxiv, 720243.

    ACKNOWLEDGE:
    This project is partially funded by CTD2, the Achilles consortium, and The Carlos Slim Foundation in
    Mexico through the Slim Initiative for Genomic Medicine.

    LICENSE: https://creativecommons.org/licenses/by/4.0/
    """
    FILE_URL = "https://ndownloader.figshare.com/files/25494359"

    @staticmethod
    def load_data(save_directory) -> AbstractDataSource:
        h5_file = os.path.join(save_directory, "achilles.h5")
        if not os.path.exists(h5_file):
            csv_file_path = os.path.join(save_directory, "achilles.csv")
            if not os.path.exists(csv_file_path):
                sp.download_streamed(Achilles.FILE_URL, csv_file_path)
            df = pd.read_csv(csv_file_path)
            gene_names = list(map(lambda x: x.split(" ")[0], df.columns.values.tolist()[1:]))
            data = df.values[:, 1:].astype(float).transpose()

            si = SimpleImputer(missing_values=float("nan"), strategy='mean')
            data = si.fit_transform(data)

            name_converter = HGNCNames(save_directory)
            gene_names = name_converter.update_outdated_gene_names(gene_names)
            gene_names, idx_start = np.unique(sorted(gene_names), return_index=True)
            data = data[idx_start]
            HDF5Tools.save_h5_file(h5_file,
                                   data,
                                   "achilles",
                                   column_names=df["DepMap_ID"].values.tolist(),
                                   row_names=gene_names)
        data_source = HDF5DataSource(h5_file, fill_missing_value=0)
        return data_source



"""
Copyright 2021 Patrick Schwab, Arash Mehrjou, GlaxoSmithKline plc; Andrew Jesson, University of Oxford; Ashkan Soleymani, MIT

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
# AUTOGENERATED! DO NOT EDIT! File to edit: 03_consistent_mc_dropout.ipynb (unless otherwise specified).

__all__ = ['BayesianModule', 'ConsistentMCDropout', 'ConsistentMCDropout2d']

# Cell

import torch
from typing import List
from torch.nn import Module

# Cell


class BayesianModule(Module):
    """A module that we can sample multiple times from given a single input batch.
    To be efficient, the module allows for a part of the forward pass to be deterministic.
    """

    k = None

    def __init__(self):
        super(BayesianModule, self).__init__()

    # Returns B x n x output
    def forward(self, input_tensors: List[torch.Tensor], k: int = 1) -> List[torch.Tensor]:
        input_B = input_tensors[0]
        BayesianModule.k = k

        mc_input_BK = BayesianModule.mc_tensor(input_B, k)
        mc_output_BK = self.mc_forward_impl(mc_input_BK)[0]
        mc_output_B_K = BayesianModule.unflatten_tensor(mc_output_BK, k)
        return [mc_output_B_K]

    def mc_forward_impl(self, mc_input_BK: torch.Tensor):
        return mc_input_BK

    @staticmethod
    def unflatten_tensor(input: torch.Tensor, k: int):
        input = input.view([-1, k] + list(input.shape[1:]))
        return input

    @staticmethod
    def flatten_tensor(mc_input: torch.Tensor):
        return mc_input.flatten(0, 1)

    @staticmethod
    def mc_tensor(input: torch.tensor, k: int):
        mc_shape = [input.shape[0], k] + list(input.shape[1:])
        return input.unsqueeze(1).expand(mc_shape).flatten(0, 1)

# Cell


class _ConsistentMCDropout(Module):
    def __init__(self, p=0.5):
        super().__init__()

        if p < 0 or p > 1:
            raise ValueError("dropout probability has to be between 0 and 1, " "but got {}".format(p))

        self.p = p
        self.mask = None

    def extra_repr(self):
        return "p={}".format(self.p)

    def reset_mask(self):
        self.mask = None

    def train(self, mode=True):
        super().train(mode)
        if not mode:
            self.reset_mask()

    def _get_sample_mask_shape(self, sample_shape):
        return sample_shape

    def _create_mask(self, input, k):
        mask_shape = [1, k] + list(self._get_sample_mask_shape(input.shape[1:]))
        mask = torch.empty(mask_shape, dtype=torch.bool, device=input.device).bernoulli_(self.p)
        return mask

    def forward(self, input: torch.Tensor):
        if self.p == 0.0:
            return input

        k = BayesianModule.k if BayesianModule.k is not None else 1
        if self.training:
            # Create a new mask on each call and for each batch element.
            k = input.shape[0]
            mask = self._create_mask(input, k)
        else:
            if self.mask is None:
                # print('recreating mask', self)
                # Recreate mask.
                self.mask = self._create_mask(input, k)

            mask = self.mask

        mc_input = BayesianModule.unflatten_tensor(input, k)
        mc_output = mc_input.masked_fill(mask, 0) / (1 - self.p)

        # Flatten MCDI, batch into one dimension again.
        return BayesianModule.flatten_tensor(mc_output)

# Cell


class ConsistentMCDropout(_ConsistentMCDropout):
    r"""Randomly zeroes some of the elements of the input
    tensor with probability :attr:`p` using samples from a Bernoulli
    distribution. The elements to zero are randomized on every forward call during training time.
    During eval time, a fixed mask is picked and kept until `reset_mask()` is called.
    This has proven to be an effective technique for regularization and
    preventing the co-adaptation of neurons as described in the paper
    `Improving neural networks by preventing co-adaptation of feature
    detectors`_ .
    Furthermore, the outputs are scaled by a factor of :math:`\frac{1}{1-p}` during
    training. This means that during evaluation the module simply computes an
    identity function.
    Args:
        p: probability of an element to be zeroed. Default: 0.5
        inplace: If set to ``True``, will do this operation in-place. Default: ``False``
    Shape:
        - Input: `Any`. Input can be of any shape
        - Output: `Same`. Output is of the same shape as input
    Examples::
        >>> m = nn.Dropout(p=0.2)
        >>> input = torch.randn(20, 16)
        >>> output = m(input)
    .. _Improving neural networks by preventing co-adaptation of feature
        detectors: https://arxiv.org/abs/1207.0580
    """
    pass


class ConsistentMCDropout2d(_ConsistentMCDropout):
    r"""Randomly zeroes whole channels of the input tensor.
    The channels to zero-out are randomized on every forward call.
    During eval time, a fixed mask is picked and kept until `reset_mask()` is called.
    Usually the input comes from :class:`nn.Conv2d` modules.
    As described in the paper
    `Efficient Object Localization Using Convolutional Networks`_ ,
    if adjacent pixels within feature maps are strongly correlated
    (as is normally the case in early convolution layers) then i.i.d. dropout
    will not regularize the activations and will otherwise just result
    in an effective learning rate decrease.
    In this case, :func:`nn.Dropout2d` will help promote independence between
    feature maps and should be used instead.
    Args:
        p (float, optional): probability of an element to be zero-ed.
        inplace (bool, optional): If set to ``True``, will do this operation
            in-place
    Shape:
        - Input: :math:`(N, C, H, W)`
        - Output: :math:`(N, C, H, W)` (same shape as input)
    Examples::
        >>> m = nn.Dropout2d(p=0.2)
        >>> input = torch.randn(20, 16, 32, 32)
        >>> output = m(input)
    .. _Efficient Object Localization Using Convolutional Networks:
       http://arxiv.org/abs/1411.4280
    """

    def _get_sample_mask_shape(self, sample_shape):
        return [sample_shape[0]] + [1] * (len(sample_shape) - 1)